#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan  8 11:06:17 2021

@author: liebig@imst.de
"""

VERSION = 0.007

import os, sys
import numpy as np
import h5py

try:
    from PyQt5.QtGui import *
    from PyQt5.QtCore import *
    from PyQt5.QtWidgets import *

    import vtk

    from vtk.qt import QVTKRWIBase
    vtk.qt.QVTKRWIBase = 'QGLWidget'

    from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
except:
    import traceback
    print('Unable to load python modules required for the GUI:')
    traceback.print_exc()

DEBUG = True  # enable verbose output

def info():
    print(f'\nsPD Data validation, version {VERSION}')
    print(   '--------------------------------------')
    print('List of authors:')
    for autor in ['Andreas Christ (andreas.christ@runbox.com)', 'Thorsten Liebig (thorsten.liebig@imst.de)']:
        print(autor)
    print()
    print( 'python version: {}.{}.{}'.format(sys.version_info.major, sys.version_info.minor, sys.version_info.micro))
    print(f'numpy version : {np.__version__}')
    print(f'h5py version  : {h5py.__version__}')

def read_sPD(fn):
    # return saved sPD data
    # V: vertex array in mm
    # F: face index array (optional, required for GUI display)
    # sPD: sPD data
    f = h5py.File(fn, 'r')
    sPD = np.array(f["spd_vertex"])
    x = np.array(f["x_axis_vertex"])
    y = np.array(f["y_axis_vertex"])
    z = np.array(f["z_axis_vertex"])
    V = np.array([x,y,z]).swapaxes(0,1)*1000
    if 'faces' in f:
        F = np.array(f["faces"])
    else:
        F = None
    meta_data = {}
    for a in ['avg', 'def', 'tot_pow']:
        if a in f.attrs:
            meta_data[a] = f.attrs[a]
    f.close()
    return V, F, sPD, meta_data

def calc_v_to_f_weights(V, F):
    # calculate the coeff to weight from vertex to face data
    w = np.zeros((V.shape[0],))
    A = 0
    for f in F:
        v0 = V[f[0]]
        v1 = V[f[1]]
        v2 = V[f[2]]
        ab = v1-v0
        ac = v2-v0
        a = 0.5*np.sqrt((ab[1]*ac[2]-ab[2]*ac[1])**2 + (ab[2]*ac[0]-ab[0]*ac[2])**2 + (ab[0]*ac[1]-ab[1]*ac[0])**2)
        A += a
        w[f] += a/3
    w *= 1e-6 # convert from mm² to m²
    A *= 1e-6
    return w, A

def validate(fn, ref_fn, max_err=None):
    print('Validate sPD data:')
    print(f'Data file: "{fn}"')
    print(f'Ref. file: "{ref_fn}"')
    print('Max. Error: {} %'.format(max_err if max_err is not None else "auto"))
    print('\nResult:')

    V1, F1, sPD1, md1 = read_sPD(fn)
    V2, F2, sPD2, md2 = read_sPD(ref_fn)
    return validate_sPD(V1, sPD1, md1, V2, sPD2, md2, max_err)

def validate_sPD(V1, sPD1, md1, V2, sPD2, md2, max_err=None):
    # validate sPD
    # V1 and V2 are assumed in mm

    dV = (V1[:,0] - V2[:,0])**2 + (V1[:,1] - V2[:,1])**2 + (V1[:,2] - V2[:,2])**2
    if np.max(dV)>1e-3:
        raise Exception('validate_sPD:: Vertex data does not match!')

    # auto detect SAR-Stars by their vertex count and set max_error to 40%
    if max_err is None and len(V1) in [52922, 51914, 51398]:
        max_err = 40
    elif max_err is None:
        max_err = 5

    status = {}
    pass_ = True
    max_val = max(np.max(sPD1), np.max(sPD2))
    diff_of_max = np.abs(np.max(sPD1) - np.max(sPD2))/max_val*100
    diff = np.abs(sPD1-sPD2)/max_val*100
    max_diff = np.max(diff)
    mean_diff = np.mean(diff)
    if diff_of_max>1:
        pass_=False
    if np.max(diff)>max_err:
        pass_=False
    o1 = len(np.where(diff>1)[0])
    o5 = len(np.where(diff>5)[0])
    o_max = len(np.where(diff>max_err)[0])
    if o5>2000:
        pass_=False

    if 'tot_pow' in md1 and 'tot_pow' in md2:
        status['max_pow_diff'] = np.abs(md1['tot_pow'] - md2['tot_pow'])/max(md1['tot_pow'], md2['tot_pow'])*100
    if DEBUG:
        print('Diff of msPD: {:>0.3f}%'.format(diff_of_max))
        print('Max  diff: {:>4.1f}%'.format(max_diff))
        print('Mean diff: {:.3f}%'.format(mean_diff))
        if 'max_pow_diff' in status:
            print('Total power diff: {:.3f}%'.format(status['max_pow_diff']))
        print('Vertex over  1%: {:>4} ({:.3f}%)'.format(o1, o1/len(diff)*100))
        print('Vertex over  5%: {:>4} ({:.3f}%)'.format(o5, o5/len(diff)*100))
        if max_err>5:
            print('Vertex over {}%: {:>4} ({:.3f}%)'.format(max_err, o_max, o_max/len(diff)*100))
        print('STATUS: ', 'PASS' if pass_ else 'FAILED')


    status['diff_of_max'] = diff_of_max
    status['max_diff'] = max_diff
    status['mean_diff'] = mean_diff
    status['count_1_over'] = o1
    status['count_5_over'] = o5

    return pass_, diff, status

MAX_DIR_DEPTH = 3
class PD_Vis_Widget(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self)
        self.setWindowTitle('sPD Validate and Compare')
        self.splitter = QSplitter(Qt.Horizontal, self)
        self.setCentralWidget(self.splitter)
        self.left = QWidget()
        vlay = QVBoxLayout()
        vlay.setContentsMargins(0,0,0,0)
        self.left.setLayout(vlay)
        hbox = QHBoxLayout()
        hbox.setContentsMargins(0,0,0,0)
        self.work_dir = QLineEdit()
        self.work_dir.editingFinished.connect(self.updateTree)
        hbox.addWidget(self.work_dir, 1)
        self.work_dir_browse = QPushButton('Browse')
        self.work_dir_browse.clicked.connect(self.browse_wd)
        hbox.addWidget(self.work_dir_browse)

        group = QGroupBox('Working Directory:')
        group.setLayout(hbox)
        vlay.addWidget(group)

        self.tree = QTreeWidget()
        self.tree.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.tree.setHeaderLabels(['Results'])
        self.tree.itemSelectionChanged.connect(self.showItem)
        vlay.addWidget(self.tree, 1)
        self.conf_wid = QGroupBox('Config && Info')
        vlay.addWidget(self.conf_wid)

        conf_lay = QFormLayout()
        self.conf_wid.setLayout(conf_lay)
        self.info_lbl = QLabel()
        conf_lay.addRow('Info:', self.info_lbl)
        self.err_lbl = QLabel()
        conf_lay.addRow('Error:', self.err_lbl)
        self.mask_val = QDoubleSpinBox()
        self.mask_val.setRange(0,100)
        self.mask_val.setSingleStep(0.1)
        self.mask_val.setDecimals(1)
        self.mask_val.setValue(0)
        self.mask_val.valueChanged.connect(self.showItem)
        conf_lay.addRow('Mask Err.:', self.mask_val)

        self.splitter.addWidget(self.left)
        self.ren_win_int = QVTKRenderWindowInteractor(self.splitter)
        self.splitter.addWidget(self.ren_win_int)
        self.ren_win_int.SetInteractorStyle(vtk.vtkInteractorStyleTrackballCamera())

        self.ren_win = self.ren_win_int.GetRenderWindow()
        self.ren = vtk.vtkRenderer()
        self.ren_win.AddRenderer(self.ren)
        self.iren = self.ren_win.GetInteractor()
        self.ren.ResetCamera()

        self.axesActor = vtk.vtkAxesActor()
        self.axes = vtk.vtkOrientationMarkerWidget()
        self.axes.SetOrientationMarker(self.axesActor)
        self.axes.SetViewport(0.0,0.0,0.3,0.3)
        self.axes.SetInteractor(self.iren)

        self.axes.EnabledOn()
        self.axes.InteractiveOff()

        self.ren_win.SetAlphaBitPlanes(1)
        self.ren_win.SetMultiSamples(0)
        self.ren.SetUseDepthPeeling(1)
        self.ren.SetMaximumNumberOfPeels(100)
        self.ren.SetOcclusionRatio(0.01)

        self.iren.Initialize()

        self.cTable = vtk.vtkLookupTable()
        self.cTable.SetNumberOfTableValues(256)
        from matplotlib import cm
        cmap = cm.get_cmap('jet')
        for n in range(256):
            self.cTable.SetTableValue(n, *cmap((n)/256.0))
        self.cTable.Build()
        self.cTable.SetNanColor(0,0,0,0)

        self.scalar_bar = vtk.vtkScalarBarActor()
        self.scalar_bar.SetOrientationToHorizontal()
        self.scalar_bar.SetLookupTable(self.cTable)

        self.scalar_bar_widget = vtk.vtkScalarBarWidget()
        self.scalar_bar_widget.SetInteractor(self.iren)
        self.scalar_bar_widget.SetScalarBarActor(self.scalar_bar)
        self.scalar_bar_widget.Off()

        self.curr_actor = None
        self.data_cache = {}

        colors = vtk.vtkNamedColors()
        # Set the background color.
        bkg = map(lambda x: x / 255.0, [26, 51, 102, 255])
        # bkg = map(lambda x: x / 255.0, [0, 0, 0, 255])
        colors.SetColor("BkgColor", *bkg)
        self.ren.SetBackground(colors.GetColor3d("BkgColor"))

        #self.updateTree()
        self.resize(1600,1000)
        self.first_ren = True
        self.tree.setFocus()


    def browse_wd(self):
        wd = self.work_dir.text()
        wd = QFileDialog.getExistingDirectory(self, 'Choose working directory', wd)
        if wd:
            self.work_dir.setText(wd)
            self.updateTree()

    def updateTree(self):
        self.clearRender()
        self.tree.clear()
        path = self.work_dir.text()
        if len(path)==0 or not os.path.exists(path):
            return
        def addItems(path, p_item, depth):
            try:
                fns = sorted(os.listdir(path))
            except PermissionError:
                return
            for fn in fns:
                if os.path.isdir(os.path.join(path, fn)):
                    dir_item = QTreeWidgetItem(p_item if p_item is not None else self.tree, [fn,], 0)
                    if p_item is None:
                        dir_item.setExpanded(True)
                    dir_item.setFlags(Qt.ItemIsEnabled)
                    if depth<=MAX_DIR_DEPTH:
                        addItems(os.path.join(path, fn), dir_item, depth+1)
                elif p_item is not None:
                    __, ext = os.path.splitext(fn)
                    if ext not in ['.sPD']:
                        continue
                    f_item = QTreeWidgetItem(p_item, [fn,], 1)
                    f_item.setData(0, Qt.UserRole, os.path.join(path, fn))
        addItems(path, None, 0)

    def clearRender(self):
        if self.curr_actor is not None:
            self.ren.RemoveActor(self.curr_actor)
            self.curr_actor = None
        self.scalar_bar_widget.Off()
        self.ren_win.Render()

    def showItem(self):
        item = self.tree.currentItem()
        if item is None:
            return
        if item.type()==0:
            return

        items = self.tree.selectedItems()
        if len(items)==2:
            # diff mode
            item = items[0]
            diff_item = items[1]
        else:
            diff_item = None

        def get_data(item):
            if item is None:
                return (None,)*4
            fn = item.data(0, Qt.UserRole )
            try:
                return read_sPD(fn)
            except:
                import tracenack
                traceback.print_exc()
                return (None,)*4
        if item is None:
            return
        fn = item.data(0, Qt.UserRole )
        V, F, data, md = get_data(item)

        self.info_lbl.setText('')
        self.err_lbl.setText('')
        self.err_lbl.setStyleSheet("")

        if V is None or data is None:
            self.clearRender()
            self.info_lbl.setText('Invalid data found!')
            return

        self.scalar_bar_widget.On()
        self.scalar_bar.SetTitle('sPD')
        V2, F2, data2, md2 = get_data(diff_item)
        if V2 is not None:
            if len(data)!=len(data2):
                self.info_lbl.setText('Data Size mismatch!')
                self.clearRender()
                return
            max_err = 20 if 'SAR' in fn else 5
            try:
                pass_, data, status = validate_sPD(V, data, md, V2, data2, md2, max_err)
            except Exception as e:
                self.info_lbl.setText('Validate failed: ' + str(e))
                self.clearRender()
                return
            mask = self.mask_val.value()
            max_diff = status['max_diff']
            avg_err  = status['mean_diff']
            info = f'max. Err: {max_diff:.1f}%, avg. Err: {avg_err:.3f}%'
            if 'max_pow_diff' in status:
                info += f", max. POW diff: {status['max_pow_diff']:.3f}%"
            if pass_:
                self.err_lbl.setText('PASS: ' + info)
                self.err_lbl.setStyleSheet("QLabel { color: green; }")
            else:
                self.err_lbl.setText('PASS: ' + info)
                self.err_lbl.setStyleSheet("QLabel { color: red; }")
            if mask>0:
                data[data<mask] = np.NaN
            self.scalar_bar.SetTitle('sPD err (%)')

        self.info_lbl.setText(f'Data Size: {len(data)}')

        if F is None:
            F = F2

        if F is None:
            self.clearRender()
            self.info_lbl.setText('No face data found, unable to display')
            return

        # render data over V and F
        points = vtk.vtkPoints()
        for v in V:
            points.InsertNextPoint(v[0], v[1], v[2])

        triangles = vtk.vtkCellArray()
        for f in F:
            triangles.InsertNextCell(3, f)

        self.trianglePolyData = vtk.vtkPolyData()
        self.trianglePolyData.SetPoints( points )
        self.trianglePolyData.SetPolys( triangles )

        from vtk.util import numpy_support
        fa = numpy_support.numpy_to_vtk(num_array=data.ravel(), deep=True, array_type=vtk.VTK_FLOAT)
        self.trianglePolyData.GetPointData().SetScalars(fa)
        self.mapper = vtk.vtkPolyDataMapper()

        self.normalGenerator = vtk.vtkPolyDataNormals()
        self.normalGenerator.SetInputData(self.trianglePolyData)
        self.normalGenerator.ComputePointNormalsOn()
        self.normalGenerator.ComputeCellNormalsOn()
        self.normalGenerator.Update()
        self.normalGenerator.ReleaseDataFlagOn()
        self.mapper.SetInputData(self.normalGenerator.GetOutput())

        self.mapper.SetScalarRange(0, np.nanmax(data))

        self.mapper.SetLookupTable(self.cTable)

        actor = vtk.vtkActor()
        actor.SetMapper(self.mapper)

        self.cTable.SetTableRange(np.nanmin(data), np.nanmax(data))

        if self.curr_actor is not None:
            self.ren.RemoveActor(self.curr_actor)

        self.ren.AddActor(actor)
        if self.curr_actor is not None:
            self.ren.ResetCameraClippingRange()

        if self.first_ren:
            self.ren.ResetCamera()
        self.first_ren = False
        self.curr_actor = actor
        self.ren_win.Render()


def main():
    if len(sys.argv)==2 and sys.argv[1]=='--gui':
        app = QApplication(sys.argv)
        wid = PD_Vis_Widget()
        wid.show()
        ec = app.exec_()
        sys.exit(ec)
    elif len(sys.argv)==2 and sys.argv[1]=='--info':
        info()
    elif len(sys.argv)in [3,4]:
        fn  = sys.argv[1]
        ref = sys.argv[2]
        max_err = 5
        if len(sys.argv)==4:
            max_err = float(sys.argv[3])
        validate(fn, ref, max_err)
    else:
        print('Usage: python validate.py ["--gui" or "--info" or sPD_data ref_data [max_err]]')
        print('       --gui    : Load GUI to validate/compare data')
        print('       --info   : Show info')
        print('       sPD_data : sPD data to validate agains <ref_data>')
        print('       ref_data : reference data file')
        print('       max_err  : max. allowed error (default auto 5%/40%)')
        print('      ')


if __name__ == '__main__':
    main()
